'
' -----[ Title ]------------------------------------------------------
'
' File......: SNAP.BAS
' Purpose...: Prototype of a S.N.A.P controlled I/O device with 64 multiplexed Bits
' Author....: Mike Eitel
' Version...: 1.20AVR small I/O device
' MCU.......: AVR 2313
' Started...: 991002
' Updated...: 991024
'
' -----[ Program Description ]----------------------------------------
'
' This program implements the S.N.A.P protocol in BASCOM for an universal 64 Bits I/O device.
' It uses STK200 development kit for testing purpouse.
' It's based on 16-bit CRC-CCITT as error detection method to give secure data transfer.
'
' This File is partialy based on a program from Christer Johansson HTH and
' I do not intent to touch any of his rights !!!
' So for correct use of his copyrights please inform as marked below!
' For more information about S.N.A.P see URL http://www.hth.com/snap/
'
' -----[ Disclaimer ]-------------------------------------------------
'
' This example is offered on an "AS IS" basis, no warranty expressed or implied.
' The programers disclaim liability of any damages associated with the use of the hardware or software
' described herein. You use it on your own risk. We are not able to provide any free support.
'
' Copyright (c) 1998-1999 HTH on theirs parts all rights reserved
' Copyright (c) 1999 Mike Eitel requests on his changed parts all rights reserved
'
' -----[ Revision History ]-------------------------------------------
'
' 980529 - Version 1.00    Initial release                                       Christer Johansson HTH
' 980919 - Version 1.01    Revised for S.N.A.P Version 0.91                      Christer Johansson HTH
' 990130 - Version 1.02    Revised for S.N.A.P Version 1.00                      Christer Johansson HTH
' 991002 - Version 1.00AVR S.N.A.P Ver 1.00 for AVR 8515                         Mike Eitel
' 991031 - Version 1.10AVR S.N.A.P Ver 1.20 small AVR 2313                       Mike Eitel
'
' -----[ Constants ]--------------------------------------------------
'
' The packet structure is defined in the received packets first two bytes (HDB2 and HDB1).
' Overview of header definition bytes (HDB2 and HDB1)
'
'                            HDB2              HDB1
'                     +-----------------+-----------------+
'                     | D D S S P P A A | C E E E N N N N |
'                     +-----------------+-----------------+
'
' The following packet structure is used.
'
'                      DD=01     - 1 Byte destination address
'                      SS=01     - 1 Byte source address
'                      PP=00     - No protocol specific flags
'                      AA=01     - Acknowledge is required
'                                         C=0/1     - Command mode not / is supported
'                                         EEE=100   - 16-bit CRC-CCITT
'                                         NNNN=0101 - 5 Byte data
'
Const Preamble = &B01010011                                   ' Preamble byte
Const Sync = &B01010100                                       ' Synchronisation byte
Const Crcpoly = &H1021                                        ' CRC-CCITT
Const Chdb2 = &B01010001                                      ' &B01010001 DD=01 - 1 Byte destination address
'                                                                          SS=01 - 1 Byte source address
'                                                                          PP=00 - No protocol specific flags
'                                                                          AA=01 - Acknowledge is required
Const Anshdb2 = &B01010000                                    ' No Ack request 01 01 00 00
Const Ackhdb2 = &B01010010                                    ' Short Ack      01 01 00 10
Const Nakhdb2 = &B01010011                                    ' Short Nak      01 01 00 11

Const Ackhdb1 = &B01000001                                    ' Short ACK / NAK  0 100 0001 with 16-bit Crc -ccitt
Const Ackhdb1s = &B00000001                                   ' Short ACK / NAK  0 000 0001 with nor CRC
Const Chdb1r = &B01000101                                     ' &B01000101 C =   1/0  - Command mode / not supported
Const Chdb1w = &B11000101                                     '            Eee = 100  - 16-bit Crc -ccitt
'                                                                          NNNN= 0101 - 5 Byte data

Const Numofdatabytes = 5                                      ' Number of transmitted Databytes

Const Myaddress = &H04                                        ' Address for this node (1-255)

Const Dataarea = 8                                            ' Size of the Databuffer

Const Timer1_reload = &H9FFF                                  ' Define fine adjust cycle of auto sending

' -----[ Variables ]--------------------------------------------------
'
Dim Temp1 As Byte                                             ' Temporary Variable
Dim Temp2 As Byte                                             ' Temporary Variable
Dim Temp_crc As Word                                          ' Temporary Variable
Dim Crc As Word                                               ' CRC word variable
Dim Crc2 As Byte                                              ' CRC hi byte variable
Dim Crc1 As Byte                                              ' CRC lo byte variable
Dim Hdb2 As Byte                                              ' Header Definition Byte 2
Dim Hdb1 As Byte                                              ' Header Definition Byte 1
Dim Sab1 As Byte                                              ' (Hisaddress) What node sent this packet
Dim Dab1 As Byte                                              ' (Myaddress) What node should have this paket
Dim Db(5) As Byte                                             ' Packet send/received command (1)
' **********************************************************************************************************
' **** ATTENTION. **** As S.N.A.P uses 5..1 and BASCOM array 1..5 numbering all ponters are changed !!! ****
' **********************************************************************************************************
Dim Hisaddress As Byte                                        ' Byte containing the last requester adress
'                            Commands
'                         +-----------------+                   W    = 0/1 Read / Write
'                         | W C C C A A A A |                   CCC  = 0. 7  Different command categories
'                         +-----------------+                        = 7 prioritised commands
'                                                                    = 6
'                                                                    = 5
'                                                                    = 4
'                                                                    = 3
'                                                                    = 2 EPROM Data Area
'                                                                    = 1 SRAM Data Area
'                                                                    = 0
'                                                               AAAA = 0..15 Pointer for values
'
Dim Commands As Byte                                          ' Byte to recognize the requested commands
Dim Datas(dataarea) As Byte                                   ' Place to keep the dynamic datas
Dim Multiplexb As Byte                                        ' Counter to multiplex 32 In- and 32 Outputpins
Dim I As Byte

' -----[ Initialization ]---------------------------------------------
'
' Define serial communication parameters
  $baud = 9600
  $crystal = 4000000

  Ddrb = &HFF                                                 ' Switch to read from multiple I/O Data port
  Ddrd = &B01111100                                           ' Set Portd Pin 2..6 as Outputs

  Config Timer0 = Timer , Prescale = 1024                     ' Define update speed
  Config Timer1 = Timer , Prescale = 256                      ' Define Autosend Raw adjust speed

  Commands = 0                                                ' Set to defined Startup value
  Multiplexb = 0                                              ' Set to defined Startup value
  Timer1 = Timer1_reload                                      ' Set to defined Startup value

  On Timer0 Timer0_isr                                        ' Define ISR when timer 0 overflows                                         '
  On Timer1 Timer1_isr                                        ' Define ISR when timer 1 overflows                                        '

  Enable Timer0                                               ' Allow cyclic multiplexing of 8 Byte latches
  Enable Interrupts                                           ' Start the system timing from this line on

'------[ Program ]----------------------------------------------------
'
'Print Chr(preamble) ; Chr(preamble) ; Myaddress ; Chr(preamble) ; Chr(preamble)

_start:

  ' Wait for data on serialport
  Inputbin Temp1

  ' If received data is SYNC byte read next seven bytes from master, if not return to start
  If Temp1 <> Sync Then
     Goto _start
  End If

  ' Get packet in binary mode
  Inputbin Hdb2 , Hdb1 , Dab1 , Sab1 , Db(1) , Crc2 , Crc1

  ' Check HDB2 to see if MCU is capable to use the packet structure, if not goto Start
  If Hdb2 <> Chdb2 Then
     Goto _start
  End If

  ' Check HDB1 to see if MCU is capable to use the packet structure, if not goto Start
  If Hdb1 = Chdb1w Then
     Commands = Db(5)                                         ' Check for used command
  Elseif Hdb1 = Chdb1r Then
     Commands = 0                                             ' Generate later NAK
  Else
     Goto _start                                              ' Start new reading cycle if not good
  End If

  ' Address check routine checks if this is the node addressed, if not goto Start
  If Dab1 <> Myaddress Then
     Goto _start
  End If

  Hisaddress = Sab1                                           'keep in mind his address

  ' CRC check routine and calculate CRC for all received bytes
  Crc = 0
  Temp2 = Hdb2
  Gosub Calc_crc
  Temp2 = Hdb1
  Gosub Calc_crc
  Temp2 = Dab1
  Gosub Calc_crc
  Temp2 = Sab1
  Gosub Calc_crc
  Temp2 = Db(1)
  Gosub Calc_crc
  Temp2 = Db(2)
  Gosub Calc_crc
  Temp2 = Db(3)
  Gosub Calc_crc
  Temp2 = Db(4)
  Gosub Calc_crc
  Temp2 = Db(5)
  Gosub Calc_crc
  Temp2 = Crc2
  Gosub Calc_crc
  Temp2 = Crc1
  Gosub Calc_crc

  ' Check if there was any CRC errors, if so, send NAK
  If Crc <> 0 Then
     Gosub _nak                                               ' Gosub _nak Routine
     Goto _start                                              ' Restart reading process
  End If

  ' No CRC errors in packet so check what to do and then start writing data or do according commands
  Select Case Commands                                        ' Decide what kind of command categorie to treat
         Case 16 :                                            ' Reading datas 1..4
                    Db(1) = Datas(1)                          ' Reading
                    Db(2) = Datas(2)                          ' Reading
                    Db(3) = Datas(3)                          ' Reading
                    Db(4) = Datas(4)                          ' Reading
                    Gosub _send                               ' Normal re-reading SRAM
         Case 17 :                                            ' Reading datas 5..8
                    Db(1) = Datas(5)                          ' Reading
                    Db(2) = Datas(6)                          ' Reading
                    Db(3) = Datas(7)                          ' reading
                    Db(4) = Datas(8)                          ' reading
                    Gosub _send                               ' Normal re-reading SRAM
         Case 144 :                                           ' Writing Datas 1..4
                    Datas(1) = Db(1)                          ' Writing
                    Datas(2) = Db(2)                          ' Writing
                    Datas(3) = Db(3)                          ' Writing
                    Datas(4) = Db(4)                          ' Writing
                    Gosub _ack                                ' End command with ACK
         Case 145 :                                           ' Writing datas 5..8
                    Datas(5) = Db(1)                          ' Writing
                    Datas(6) = Db(2)                          ' Writing
                    Datas(7) = Db(3)                          ' Writing
                    Datas(8) = Db(4)                          ' Writing
                    Gosub _ack                                ' End command with ACK
         Case 254 :                                           ' Spezial / priorised commands
                    Disable Timer1                            ' Disallow cyclic send of Datas 1..4
                    Gosub _ack
         Case 255 :
                    Enable Timer1                             ' Allow cyclic send of Datas 1..4
                    Gosub _ack                                ' Command was sucessfull
         Case Else :
                    Gosub _nak                                ' NOT recognized command
  End Select

  ' Done, go back to Start and wait for a new packet
  Goto _start                                                 ' Wait for the next telegramm


' End of the main program
' -----[ End ]--------------------------------------------------------



' -----[ Subroutines ]------------------------------------------------
'
    ' Subroutine for calculating CRC value in variable Temp2
Calc_crc:
    Temp_crc = Temp2
    Shift Temp_crc , Left , 8                                 ' Smaller as Temp_crc = Temp2 * 256

    Crc = Temp_crc Xor Crc
    For Temp1 = 0 To 7
        If Crc > 32767 Then
           Shift Crc , Left                                   ' Smaller as Crc = Crc * 2
           Crc = Crc Xor Crcpoly
        Else
           Shift Crc , Left                                   ' Smaller as Crc = Crc * 2
        End If
        Crc = Crc And 65535
    Next
Return


  ' Send NAK (i.e tell master that packet was bad)
_nak:
    Hdb2 = Nakhdb2                                            ' Set ACK bits to NAK in HDB2 (xxxxxx11)
    Goto _ret
    ' Send Ack (i.e Tell Master That Packet was Ok)
_ack:
    Hdb2 = Ackhdb2                                            ' Set ACK bits to ACK in HDB2 (xxxxxx10)
_ret:
    Sab1 = Myaddress
    Dab1 = Hisaddress
    '      Db(5) = 0                                          ' Return for ACK / NAK has no value in db1
    '      Crc = 0                                            ' Clear CRC variable
    '      Temp2 = Hdb2
    '      Gosub Calc_crc
    '      Temp2 = Ackhdb1
    '      Gosub Calc_crc
    '      Temp2 = Dab1        ****** TO SPARE CODE AND TIME NO CRC USED FOR ACK / NAK *****
    '      Gosub Calc_crc
    '      Temp2 = Sab1
    '      Gosub Calc_crc
    '      Temp2 = 0
    '      Gosub Calc_crc
    '      Crc2 = High(crc)
    '      Crc1 = Low(crc)
    '      Print Chr(sync) ; Chr(hdb2) ; Chr(ackhdb1) ; Chr(dab1) ; Chr(sab1) ; Chr(0) ; Chr(crc2) ; Chr(crc1);
    Print Chr(sync) ; Chr(hdb2) ; Chr(ackhdb1s) ; Chr(dab1) ; Chr(sab1) ; Chr(0) ;
Return


' Subroutine for sending values
_send:
    Dab1 = Hisaddress                                         ' Change SAB1 and DAB1 address bytes
    Sab1 = Myaddress                                          ' Change SAB1 and DAB1 address bytes
    Crc = 0                                                   ' Clear CRC variable

    Temp2 = Anshdb2                                           ' HDB2. Always answering with defined format
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Chdb1r                                            ' HDB1. Always answering with read flag
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Dab1                                              ' Put DAB1 in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Sab1                                              ' Put SAB1 in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Db(1)                                             ' Put DATA Byte 1 in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Db(2)                                             ' Put DATA Byte 2 in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Db(3)                                             ' Put DATA Byte 3 in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Db(4)                                             ' Put DATA Byte 4 in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC
    Temp2 = Db(5)                                             ' Put Command Byte in tempory variable
    Gosub Calc_crc                                            ' Calculate CRC

    Crc2 = High(crc)                                          ' Put calculated Hi_CRC value in CRC2
    Crc1 = Low(crc)                                           ' Put calculated Lo_CRC value in CRC1

    ' Send packet to master, including the SYNC byte
'    ******** CRC in S.N.A.P produces to much errors when data between 128 and 160 !!!  *******
'    Print Chr(preamble) ; Chr(sync) ; Chr(anshdb2) ; Chr(chdb1r) ; Chr(dab1) ; Chr(sab1) ; Chr(sdb(5)) ; Chr(sdb(4)) ; Chr(sdb(3)) ; Chr(sdb(2)) ; Chr(sdb(1)) ; Chr(crc2) ; Chr(crc1)
'
    Print Chr(sync) ; Chr(anshdb2) ; Chr(&B00000101) ; Chr(dab1) ; Chr(sab1) ; Chr(db(1)) ; Chr(db(2)) ; Chr(db(3)) ; Chr(db(4)) ; ; Chr(db(5)) ;
    Return                                                    ' End of send function


Timer0_isr:                                                   ' Controlls the multiplexing of data to I/O's
    Incr Multiplexb                                           ' Cycle counter for the datas

    Select Case Multiplexb                                    ' From 1 ..8
            Case 1 :
' ***** Only for testing. Normaly DDRB is &H00 for reading !!! *******
' ***** Also normal instruction is reading !!! *******
' *****                   Ddrb = &H00                               ' Switch to read from multiple I/O Data port
                    Ddrb = &HFF                               ' Switch to READ from multiple I/O Data port
                    Portd = &B01111000                        ' Read from Latch 1
                    Portb = Datas(multiplexb)                 '
' *****                   Datas(multiplexb) = Portb                 ' Read the data from I/O
           Case 2 :
                    Portd = &B01110100                        ' Read from Latch 2
                    Portb = Datas(multiplexb)                 ' Read the data from I/O
' *****                   Datas(multiplexb) = Portb                 ' Read the data from I/O
           Case 3 :
                    Portd = &B01101100                        ' Read from Latch 3
                    Portb = Datas(multiplexb)                 ' Read the data from I/O
' *****                   Datas(multiplexb) = Portb                 ' Read the data from I/O
           Case 4 :
                    Portd = &B01011100                        ' Read from Latch 4
                    Portb = Datas(multiplexb)                 ' Read the data from I/O
' *****                   Datas(multiplexb) = Portb                 ' Read the data from I/O
           Case 5 :
                    Ddrb = &HFF                               ' Switch to WRITE to multiple I/O Data port                    Portb = Datas(multiplexb)                 ' Transfer the data
                    Portd = &B00111000                        ' Write to Latch 5
                    Portb = Datas(multiplexb)                 ' Transfer the data
           Case 6 :
                    Portd = &B00110100                        ' Write to Latch 6
                    Portb = Datas(multiplexb)                 ' Transfer the data
           Case 7 :
                    Portd = &B00101100                        ' Write to Latch 7
                    Portb = Datas(multiplexb)                 ' Transfer the data
           Case 8 :
                    Portd = &B00011100                        ' Write to Latch 8
                    Portb = Datas(multiplexb)                 ' Transfer the data
                    Multiplexb = 0                            ' Restart the multiplexer process
    End Select
Return

Timer1_isr:
    Timer1 = Timer1_reload
    Db(5) = Db(5) - 1                                         ' ***** TEST Modus for S.N.A.P DLL error *********
'    Db(5) = 16                                                ' Auto send the 32 InputBits
    Db(4) = Datas(1)                                          ' Auto send back read data
    Db(3) = Datas(2)                                          ' Auto send back read data
    Db(2) = Datas(3)                                          ' Auto send back read data
    Db(1) = Datas(4)                                          ' Auto send back read data

    Gosub _send
Return